// quickselect.cpp
// Glenn G. Chappell
// 28 Sep 2012
//
// For CS 411/611 Fall 2012
// Lomuto Partition & Quickselect

#include <iostream>
using std::cout;
using std::endl;
using std::cin;
#include <algorithm>
using std::iter_swap;
#include <cstdlib>
using std::size_t;
#include <cassert>  // for assert
#include <vector>
using std::vector;


// lpartition
// Partitions a sequence about a given pivot. Uses Lomuto Algorithm.
// Returns the new pivot position.
// 
// Requirements on Types:
//     FDIter must be a forward iterator type.
//     operator< must be a total order on the value type of FDIter.
// Pre:
//     [first, last) is a valid nonempty range.
//     pivotiter is an iterator in the range [first, last).
// Post:
//     [first, last) contains the same items as it did initially,
//      but now each ITEM before pivotiter has ITEM < *pivotiter,
//      and each ITEM after pivotiter has !(ITEM < *pivotiter).
//     pivotiter lies in [first, last) and references an item with
//      the same value as *pivotiter before the function call.
template <typename FDIter>
void lpartition(FDIter first, FDIter last, FDIter & pivotiter)
{
    // Put the pivot at the start of the list
    iter_swap(first, pivotiter);

    // Now "first" points to the pivot

    // Make the "left list": list of items less than pivot
    FDIter leftFinal = first;  // points to final item in left list
    FDIter check = first;      // item to check
    for (++check;              // start after pivot, iterate through list
         check != last;
         ++check)
    {
        if (*check < *first)              // if item < PIVOT
        {
            ++leftFinal;                  // move up end mark
            iter_swap(leftFinal, check);  // and put item in left list
        }
    }

    // Note new pivot position, for caller, and put pivot there
    pivotiter = leftFinal;
    iter_swap(first, pivotiter);
}


// quickselect
// Selection using Quickselect algorithm. [first, last) specifies a
//  range. index is an index in this range (0 .. size-1). Return value
//  is an iterator to the item in this range that would be at the given
//  index, if the range were sorted (not necessarily in a stable
//  manner). The range is modified, but not sorted.
// Uses lpartition.
// Pre:
//     [first, last) is a valid range.
//     0 <= index < (last-first).
// Post:
//     Return is iterator to the item that would be at position given by
//      index, if the data were sorted (not necessarily in a stable
//      manner).
//     Data in [first, last) are same items as initially, but possibly
//      in a different order.
template <typename RAIter>
RAIter quickselect(RAIter first, RAIter last, size_t index)
{
    while (true)  // For tail-recursion elimination
    {
        RAIter pivotiter;  // Iterator to item chosen as pivot
    
        // Find median-of-three and point pivotiter at it.
        // We do the same comparisons as an Insertion Sort
        // of a 3-item list, but we do not modify the list.
        size_t size = last - first;
        RAIter middle = first + size/2;
        RAIter finalitem = last-1;
        if (*middle < *first)
        {
            if (*finalitem < *first)
                pivotiter = (*finalitem < *middle) ? middle : finalitem;
            else
                pivotiter = first;
        }
        else
        {
            if (*finalitem < *middle)
                pivotiter = (*finalitem < *first) ? first : finalitem;
            else
                pivotiter = middle;
        }
    
        // Do partition
        lpartition(first, last, pivotiter);
        size_t pivotindex = pivotiter - first;
    
        // BASE CASE of former recursive function
        
        // If the pivot is our item, then return it
        if (index == pivotindex)
            return pivotiter;
    
        // RECURSIVE CASE of former recursive function
        
        if (index < pivotindex)
        {   // Recurse on range of items preceding the pivot
            last = pivotiter;
            return quickselect(first, pivotiter, index);
        }
        else
        {   // Recurse on range of items following the pivot
            index -= pivotindex+1;
            first = pivotiter+1;
        }
        // Tail call is gone, replaced by loop
        // return quickselect(first, last, index);
    }
}


// Main program
// Demonstrates use of function quickselect
int main()
{
    // Set up array & copy contents to a vector
    const int SIZE = 9;
    int data[SIZE] = { 2,5,3,7,1,8,9,4,6 };
    vector<int> v(data, data+SIZE);

    // Print unaltered data
    cout << "Initial data: ";
    for (size_t i = 0; i != v.size(); ++i)
        cout << v[i] << " ";
    cout << endl;

    // Do the selection
    size_t index = 4;  // Index to select
    vector<int>::iterator pos = quickselect(v.begin(), v.end(), index);

    // Print results
    cout << "Revised data:   ";
    for (size_t i = 0; i != v.size(); ++i)
        cout << v[i] << " ";
    cout << endl;
    cout << endl;
    cout << "Value at index " << index << " in sorted data: "
         << *pos << endl;
    cout << endl;

    cout << "Press ENTER to quit ";
    while (cin.get() != '\n') ;

    return 0;
}

